from __future__ import annotations

from django.db import models
from model_utils import FieldTracker

from bulk_tracker.helper_objects import TrackingInfo
from bulk_tracker.managers import BulkTrackerManager
from bulk_tracker.signals import (
    send_post_create_signal,
    send_post_delete_signal,
    send_post_update_signal,
)


class BulkTrackerModel(models.Model):
    objects = BulkTrackerManager()
    tracker: FieldTracker

    class Meta:
        abstract = True

    def _save_table(self, *args, **kwargs):
        updated = super()._save_table(*args, **kwargs)
        self._is_created = not updated
        return updated

    def save(self, tracking_info_: TrackingInfo | None = None, **kwargs):
        if hasattr(self, "tracker") and self.tracker:
            changed = self.tracker.changed()
            super().save(**kwargs)
            if not self._is_created and changed:
                send_post_update_signal(
                    [self], model=self.__class__, old_values={self.pk: changed}, tracking_info_=tracking_info_
                )
            elif self._is_created:
                send_post_create_signal([self], model=self.__class__, tracking_info_=tracking_info_)

        else:
            raise AttributeError(
                f"Model {self.__class__} doesn't have tracker, please add `tracker = FieldTracker()` to your model"
            )

    def delete(self, *args, tracking_info_: TrackingInfo | None = None, **kwargs):
        send_post_delete_signal([self], model=self.__class__, tracking_info_=tracking_info_)
        return super().delete(*args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar

from django.db import models
from django.db.models import Model


_T = TypeVar("_T", bound=models.Model)
User = TypeVar("User", bound=models.Model)


@dataclass
class ModifiedObject(Generic[_T]):
    instance: _T
    changed_values: dict[str, Any]


@dataclass
class TrackingInfo:
    user: User | None = None
    comment: str | None = None
    reason: Model | None = None
    system: str | None = None
    kwargs: dict[str, Any] = field(default_factory=dict)
    is_robust: bool = False



from __future__ import annotations

from datetime import datetime
from unittest.mock import patch

from django.test import TransactionTestCase

from bulk_tracker.helper_objects import ModifiedObject, TrackingInfo
from bulk_tracker.signals import post_delete_signal
from tests.models import Author, Post


class TestDeleteSignal(TransactionTestCase):
    def setUp(self):
        self.author_john = Author.objects.create(first_name="John", last_name="Doe")

    def test_model_delete_should_emit_post_delete_signal(self):
        # Arrange
        signal_called_with = {}

        def post_delete_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            signal_called_with["objects"] = objects
            signal_called_with["tracking_info_"] = tracking_info_

        post_delete_signal.connect(post_delete_receiver, sender=Post)
        post = Post.objects.create(title="Sound of Winter", publish_date="1998-03-08", author=self.author_john)

        # Act
        post.delete()
        # Assert
        self.assertTrue(signal_called_with != {})
        self.assertEqual(1, len(signal_called_with["objects"]))
        self.assertEqual(None, signal_called_with["tracking_info_"])

        modified_objects: list[ModifiedObject[Post]] = signal_called_with["objects"]
        self.assertEqual("Sound of Winter", modified_objects[0].instance.title)
        self.assertEqual("1998-03-08", modified_objects[0].instance.publish_date)
        self.assertEqual(self.author_john, modified_objects[0].instance.author)
        self.assertEqual({}, modified_objects[0].changed_values)

    def test_query_delete_should_emit_post_delete_signal(self):
        # Arrange
        signal_called_with = {}

        def post_delete_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            signal_called_with["objects"] = objects
            signal_called_with["tracking_info_"] = tracking_info_

        post_delete_signal.connect(post_delete_receiver, sender=Post)
        posts = [
            Post(title="Sound of Winter", publish_date="1998-01-08", author=self.author_john),
            Post(title="Sound of Summer", publish_date="1998-06-08", author=self.author_john),
        ]
        Post.objects.bulk_create(posts)

        # Act
        Post.objects.all().order_by("publish_date").delete()

        # Assert
        self.assertTrue(signal_called_with != {})
        self.assertEqual(2, len(signal_called_with["objects"]))
        self.assertEqual(None, signal_called_with["tracking_info_"])

        modified_objects: list[ModifiedObject[Post]] = signal_called_with["objects"]

        self.assertEqual("Sound of Winter", modified_objects[0].instance.title)
        self.assertEqual(datetime.strptime("1998-01-08", "%Y-%m-%d").date(), modified_objects[0].instance.publish_date)
        self.assertEqual(self.author_john, modified_objects[0].instance.author)
        self.assertEqual({}, modified_objects[0].changed_values)

        self.assertEqual("Sound of Summer", modified_objects[1].instance.title)
        self.assertEqual(datetime.strptime("1998-06-08", "%Y-%m-%d").date(), modified_objects[1].instance.publish_date)
        self.assertEqual(self.author_john, modified_objects[1].instance.author)
        self.assertEqual({}, modified_objects[1].changed_values)

    @patch("bulk_tracker.signals.post_delete_signal.send")
    def test_delete_signal_should_be_only_emitted_after_transaction_commit(self, mocked_signal):
        # Arrange

        def post_delete_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            pass

        post_delete_signal.connect(post_delete_receiver, sender=Post)

        # Act
        from django.db import transaction

        with transaction.atomic():
            post = Post.objects.create(title="Sound of Winter", publish_date="1998-01-08", author=self.author_john)
            post.delete()
            mocked_signal.assert_not_called()

            Author.objects.create(first_name="Jane", last_name="Doe")

        # Assert
        mocked_signal.assert_called_once()
        signal_called_with_first = mocked_signal.call_args_list[0].kwargs
        self.assertEqual(signal_called_with_first["sender"], Post)

    @patch("bulk_tracker.signals.post_delete_signal.send")
    @patch("bulk_tracker.signals.post_delete_signal.send_robust")
    def test_should_use_robust_send_if_is_robust_is_true_in_tracking_info(self, mocked_signal_robust, mocked_signal):
        # Arrange

        # Act
        self.author_john.delete(tracking_info_=TrackingInfo(is_robust=True))

        # Assert
        mocked_signal.assert_not_called()
        mocked_signal_robust.assert_called_once()



INSTALLED_APPS = (
    "bulk_tracker",
    "tests",
)
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    },
}
SECRET_KEY = "dummy"

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

USE_TZ = False



from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

from django.db import transaction
from django.dispatch import Signal

from bulk_tracker.helper_objects import ModifiedObject, TrackingInfo


if TYPE_CHECKING:
    from bulk_tracker.models import BulkTrackerModel


"""
Signals that will be emitted when a bulk operations are used.
If you use the BulkTrackerModel
"""

"""
The receiver function should look like this
@receiver(post_update_signal, sender=MyModel)
def i_am_a_receiver_function(
    sender,
    objects: list[ModifiedObject[MyModel]],
    tracking_info_: TrackingInfo | None = None,
    **kwargs,
):
    do_stuff()
"""
post_update_signal = Signal()  # custom signal for bulk and single update


"""
@receiver(post_create_signal, sender=MyModel)
def i_am_a_receiver_function(
    sender,
    objects: list[ModifiedObject[MyModel]],
    tracking_info_: TrackingInfo | None = None,
    **kwargs,
):
    do_stuff()
"""
post_create_signal = Signal()  # custom signal for bulk and single create

"""
@receiver(post_delete_signal, sender=MyModel)
def i_am_a_receiver_function(
    sender,
    objects: list[ModifiedObject[MyModel]],
    tracking_info_: TrackingInfo | None = None,
    **kwargs,
):
    do_stuff()
"""
post_delete_signal = Signal()  # custom signal for bulk and single delete


def send_post_create_signal(
    objs: Iterable[BulkTrackerModel], model: type[BulkTrackerModel], tracking_info_: TrackingInfo | None = None
):
    modified_objects = [ModifiedObject(ob, {}) for ob in objs]
    if modified_objects:
        if tracking_info_ and tracking_info_.is_robust:
            method = post_create_signal.send_robust
        else:
            method = post_create_signal.send
        transaction.on_commit(
            lambda: method(
                objects=modified_objects,
                sender=model,
                tracking_info_=tracking_info_,
            )
        )


def send_post_update_signal(
    queryset: Iterable[BulkTrackerModel],
    model: type[BulkTrackerModel],
    old_values: dict[int, [dict[str, Any]]],  # {pk: changed_values}
    tracking_info_: TrackingInfo | None = None,
) -> None:
    modified_objects = []
    for obj in queryset:
        changed = old_values[obj.pk]
        diff_dict = {}
        for key, old_value in changed.items():
            if getattr(obj, key) != old_value:  # if new_values != old_value
                diff_dict[key] = old_value
        if diff_dict:
            modified_objects.append(ModifiedObject(obj, diff_dict))

    if modified_objects:
        if tracking_info_ and tracking_info_.is_robust:
            method = post_update_signal.send_robust
        else:
            method = post_update_signal.send
        transaction.on_commit(
            lambda: method(
                sender=model,
                objects=modified_objects,
                tracking_info_=tracking_info_,
            )
        )


def send_post_delete_signal(
    objs: Iterable[BulkTrackerModel],
    model: type[BulkTrackerModel],
    tracking_info_: TrackingInfo | None = None,
):
    modified_objects = [ModifiedObject(ob, {}) for ob in objs]
    if modified_objects:
        if tracking_info_ and tracking_info_.is_robust:
            method = post_delete_signal.send_robust
        else:
            method = post_delete_signal.send
        transaction.on_commit(
            lambda: method(
                objects=modified_objects,
                sender=model,
                tracking_info_=tracking_info_,
            )
        )



from __future__ import annotations

from django.db import connections, transaction
from django.db.models import Expression, Manager, QuerySet
from django.db.models.expressions import Case, Value, When
from django.db.models.functions import Cast

from bulk_tracker.helper_objects import TrackingInfo
from bulk_tracker.signals import (
    post_delete_signal,
    post_update_signal,
    send_post_create_signal,
    send_post_delete_signal,
    send_post_update_signal,
)
from bulk_tracker.utils import get_old_values


class BulkTrackerQuerySet(QuerySet):
    def update(self, *, tracking_info_: TrackingInfo | None = None, **kwargs):
        """
        This will be the single place for updating objects
        This will prevent other users from using post_save signals and other built-in signals
        because in my opinion they are bad practice, and prevent developers from supporting bulk operations
        instead this will send `post_update_signal` with all the changed objects and their old_values

        if `post_update_signal` has listeners this will result in an extra 2 queries in order to retrieve the diff.
        """
        signal_has_listener = post_update_signal.has_listeners(sender=self.model)
        # if the model doesn't have any listener on this signal, don't bother doing anything
        if not signal_has_listener:
            return super().update(**kwargs)

        # if we have listeners:
        # 1- we will consume the queryset
        old_values = {}
        pks = []
        for obj in self:
            pks.append(obj.pk)
            old_values[obj.pk] = get_old_values(obj, kwargs)

        # 2- create a new queryset based on the PK.
        # because the user may be updating the same value as the criteria which will lead to an empty queryset if we
        # loop on `self` again. i.e. `Post.objects.filter(title="The Midnight Wolf").update(title="The Sunset Wolf")`
        queryset = self.model.objects.filter(pk__in=pks).only(*kwargs.keys())

        result = super().update(**kwargs)
        send_post_update_signal(queryset, self.model, old_values, tracking_info_)
        return result

    def create(self, *, tracking_info_: TrackingInfo | None = None, **kwargs):
        """
        Create a new object with the given kwargs, saving it to the database
        and returning the created object.
        """

        # NOTE: send_post_create_signal is called after 'create' gets executed, where model.save is called
        obj = super().create(**kwargs)
        return obj

    # This function is just copied from django core, with minor modification to accept TrackingInfo
    def bulk_update(
        self, objs, fields, batch_size=None, *args, tracking_info_: TrackingInfo | None = None, **kwargs
    ) -> None:
        """
        Update the given fields in each of the given objects in the database.
        """
        if batch_size is not None and batch_size < 0:
            raise ValueError("Batch size must be a positive integer.")
        if not fields:
            raise ValueError("Field names must be given to bulk_update().")
        objs = tuple(objs)
        if any(obj.pk is None for obj in objs):
            raise ValueError("All bulk_update() objects must have a primary key set.")
        fields = [self.model._meta.get_field(name) for name in fields]
        if any(not f.concrete or f.many_to_many for f in fields):
            raise ValueError("bulk_update() can only be used with concrete fields.")
        if any(f.primary_key for f in fields):
            raise ValueError("bulk_update() cannot be used with primary key fields.")
        if not objs:
            return
        # PK is used twice in the resulting update query, once in the filter
        # and once in the WHEN. Each field will also have one CAST.
        max_batch_size = connections[self.db].ops.bulk_batch_size(["pk", "pk"] + fields, objs)
        batch_size = min(batch_size, max_batch_size) if batch_size else max_batch_size
        requires_casting = connections[self.db].features.requires_casted_case_in_updates
        batches = (objs[i : i + batch_size] for i in range(0, len(objs), batch_size))
        updates = []
        for batch_objs in batches:
            update_kwargs = {}
            for field in fields:
                when_statements = []
                for obj in batch_objs:
                    attr = getattr(obj, field.attname)
                    if not isinstance(attr, Expression):
                        attr = Value(attr, output_field=field)
                    when_statements.append(When(pk=obj.pk, then=attr))
                case_statement = Case(*when_statements, output_field=field)
                if requires_casting:
                    case_statement = Cast(case_statement, output_field=field)
                update_kwargs[field.attname] = case_statement
            updates.append(([obj.pk for obj in batch_objs], update_kwargs))
        with transaction.atomic(using=self.db, savepoint=False):
            for pks, update_kwargs in updates:
                self.filter(pk__in=pks).update(tracking_info_=tracking_info_, **update_kwargs)

    def bulk_create(self, *args, tracking_info_: TrackingInfo | None = None, **kwargs):
        """
        Insert each of the instances into the database. Do *not* call
        save() on each of the instances, do not send any pre/post_save
        signals, and do not set the primary key attribute if it is an
        autoincrement field (except if features.can_return_rows_from_bulk_insert=True).
        Multi-table models are not supported.
        Will send `post_create_signal` with the created objects
        """
        objs = super().bulk_create(*args, **kwargs)
        send_post_create_signal(objs, self.model, tracking_info_)
        return objs

    def delete(self, *, tracking_info_: TrackingInfo | None = None, **kwarg):
        if not post_delete_signal.has_listeners(self.model):
            return super().delete()
        objs = (obj for obj in self)
        result = super().delete()
        send_post_delete_signal(objs, self.model, tracking_info_)
        return result


class BulkTrackerManager(Manager.from_queryset(BulkTrackerQuerySet)):
    pass



from __future__ import annotations

from datetime import datetime
from unittest.mock import patch

from django.test import TransactionTestCase

from bulk_tracker.helper_objects import ModifiedObject, TrackingInfo
from bulk_tracker.signals import post_update_signal
from tests.models import Author, Post


class TestUpdateSignal(TransactionTestCase):
    def setUp(self):
        self.author_john = Author.objects.create(first_name="John", last_name="Doe")
        self.author_soha = Author.objects.create(first_name="Soha", last_name="Reid")

        Post.objects.create(title="Defend the Lie", publish_date="1999-05-19", author=self.author_john)
        Post.objects.create(title="Cold Vice", publish_date="2001-07-22", author=self.author_john)
        Post.objects.create(title="Sound of Winter", publish_date="2000-09-12", author=self.author_john)

        Post.objects.create(title="Prince's Advent", publish_date="1999-05-19", author=self.author_soha)
        Post.objects.create(title="The Midnight Wolf", publish_date="2002-01-12", author=self.author_soha)

    @staticmethod
    def use_bulk_update():
        posts = Post.objects.filter(publish_date="1999-05-19").order_by("publish_date")
        for post in posts:
            post.publish_date = "1999-12-31"
        signal_called_with: dict = {}
        return posts, signal_called_with

    def test_bulk_update_should_emit_post_update_signal(self):
        # Arrange
        signal_called_with: dict = {}

        def post_update_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            signal_called_with["objects"] = objects
            signal_called_with["tracking_info_"] = tracking_info_

        post_update_signal.connect(post_update_receiver, sender=Post)

        # Act
        posts = Post.objects.filter(publish_date="1999-05-19").order_by("publish_date")
        for post in posts:
            post.publish_date = "1999-12-31"
        Post.objects.bulk_update(posts, fields=["publish_date"])

        # Assert
        self.assertTrue(signal_called_with != {})
        self.assertEqual(2, len(signal_called_with["objects"]))
        self.assertEqual(None, signal_called_with["tracking_info_"])

        ### split into another test case
        modified_objects: list[ModifiedObject[Post]] = signal_called_with["objects"]
        for modified_object in modified_objects:
            self.assertEqual(datetime.strptime("1999-12-31", "%Y-%m-%d").date(), modified_object.instance.publish_date)
            self.assertEqual(
                datetime.strptime("1999-05-19", "%Y-%m-%d").date(), modified_object.changed_values["publish_date"]
            )

    def test_should_raise_attribute_error_when_model_does_not_have_tracker(self):
        # Arrange
        cache_tracker = Post.tracker
        del Post.tracker
        posts = Post.objects.filter(publish_date="1999-05-19").order_by("publish_date")
        for post in posts:
            post.publish_date = "1999-12-31"

        # Act & Assert
        with self.assertRaises(AttributeError) as context:
            posts[0].save()
        self.assertEqual(
            f"Model {Post} doesn't have tracker, please add `tracker = FieldTracker()` to your model",
            context.exception.args[0],
        )

        # Cleanup
        Post.tracker = cache_tracker

    def test_model_save_should_emit_post_update_signal(self):
        # Arrange
        signal_called_with = {}

        def post_update_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            signal_called_with["objects"] = objects
            signal_called_with["tracking_info_"] = tracking_info_

        post_update_signal.connect(post_update_receiver, sender=Post)

        # Act
        post = Post.objects.get(title="Sound of Winter")
        post.title = "Sound of Summer"
        post.save()

        # Assert
        self.assertTrue(signal_called_with != {})
        self.assertEqual(1, len(signal_called_with["objects"]))
        self.assertEqual(None, signal_called_with["tracking_info_"])

        ### split into another test case
        modified_objects: list[ModifiedObject[Post]] = signal_called_with["objects"]
        self.assertEqual("Sound of Summer", modified_objects[0].instance.title)
        self.assertEqual("Sound of Winter", modified_objects[0].changed_values["title"])

    def test_queryset_update_should_emit_post_update_signal(self):
        # Arrange
        signal_called_with = {}

        def post_update_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            signal_called_with["objects"] = objects
            signal_called_with["tracking_info_"] = tracking_info_

        post_update_signal.connect(post_update_receiver, sender=Post)

        # Act
        Post.objects.filter(title="The Midnight Wolf").update(title="The Sunset Wolf")

        # Assert
        self.assertTrue(signal_called_with != {})
        self.assertEqual(1, len(signal_called_with["objects"]))
        self.assertEqual(None, signal_called_with["tracking_info_"])

        modified_objects: list[ModifiedObject[Post]] = signal_called_with["objects"]
        self.assertEqual("The Sunset Wolf", modified_objects[0].instance.title)
        self.assertEqual("The Midnight Wolf", modified_objects[0].changed_values["title"])

    @patch("bulk_tracker.signals.post_update_signal.send")
    def test_update_signal_should_be_only_emitted_after_transaction_commit(self, mocked_signal):
        # Arrange

        def post_update_receiver(
            sender,
            objects: list[ModifiedObject[Post]],
            tracking_info_: TrackingInfo | None = None,
            **kwargs,
        ):
            pass

        post_update_signal.connect(post_update_receiver, sender=Post)
        post = Post.objects.create(title="Sound of Summer", publish_date="1998-06-08", author=self.author_john)

        # Act
        from django.db import transaction

        with transaction.atomic():
            post.title = "Sound of Winter"
            Post.objects.bulk_update([post], fields=["title"])
            mocked_signal.assert_not_called()

            self.author_john.first_name = "Johny"
            Author.objects.bulk_update([self.author_john], fields=["first_name"])

        # Assert
        mocked_signal.assert_called_once()
        signal_called_with_first = mocked_signal.call_args_list[0].kwargs
        self.assertEqual(signal_called_with_first["sender"], Post)

    @patch("bulk_tracker.signals.post_update_signal.send")
    @patch("bulk_tracker.signals.post_update_signal.send_robust")
    def test_should_use_robust_send_if_is_robust_is_true_in_tracking_info(self, mocked_signal_robust, mocked_signal):
        # Arrange
        self.author_john.first_name = "Johny"

        # Act
        self.author_john.save(tracking_info_=TrackingInfo(is_robust=True))

        # Assert
        mocked_signal.assert_not_called()
        mocked_signal_robust.assert_called_once()



from django.db import models
from model_utils import FieldTracker

from bulk_tracker.models import BulkTrackerModel


class Author(BulkTrackerModel):
    first_name = models.CharField(max_length=200)
    last_name = models.CharField(max_length=200)

    tracker = FieldTracker()


class Post(BulkTrackerModel):
    title = models.CharField(max_length=255)
    publish_date = models.DateField()
    author = models.ForeignKey(Author, on_delete=models.CASCADE, related_name="posts")

    tracker = FieldTracker()
